;; SIMUTRANS TOOLS FOR GIMP 0.5
;;
;; ver. 0.5 - 15/03/2012
;; ADD Lookup (convert) operation in Special Color Helper
;; ADD Lighten/Darken operation in Special Color Helper
;; ADD Special Color Helper sample merged select
;; ADD Set grid tool (optional resize)
;; ADD Swap colors tool
;; ADD internal stats, selection-save, selection-by-color, selection-fill 
;;     and color-lookup functions
;; CHG Export flatten now optional (saves time for large images if not needed)
;; CHG code optimization
;;
;; ver. 0.4 - 13/03/2012
;; ADD PNG Export tool
;; FIX Remove and Repair only worked on first open image (image id=1)
;;
;; ver. 0.2 - 09/03/2012
;; FIX it painted all screen if the color was not found
;; CHG some optimization 
;; CHG progress bar now working
;;
;; ver. 0.1 - 08/03/2012
;; initial release
;;
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the Simutrans Artistic License.
;; See http://forum.simutrans.com
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  



(define (selection-save image) 
    (if (zero? (car (gimp-selection-is-empty image))) (car (gimp-selection-save image)) 0))


    
(define (selection-by-color 
    drawable
    color
    threshold
    selection-mask
    )
    (gimp-by-color-select drawable color threshold CHANNEL-OP-REPLACE FALSE FALSE 0 FALSE)
    (unless (zero? selection-mask) 
        (gimp-selection-combine selection-mask CHANNEL-OP-INTERSECT)
        )
    )



(define (selection-fill
    drawable
    color
    )
    (when (zero? (car (gimp-selection-is-empty (car (gimp-drawable-get-image drawable)))))
        (gimp-context-set-foreground color)
        (gimp-edit-fill drawable FOREGROUND-FILL)
        )
    )
    
    
   
(define (color-lookup image color vertical)
    (let* (
        (drawable (car (gimp-image-get-active-layer image)))
        (color1 '()) 
        (color2 '()) 
        (koord  '())
        (x1 0) (y1 0)
        (x2 0) (y2 0)
        )
        (gimp-by-color-select drawable color 0 CHANNEL-OP-REPLACE FALSE FALSE 0 FALSE)
        (set! koord (gimp-selection-bounds image))
        (gimp-selection-none image)
        (if (zero? (car koord))
            '()
            (begin
                (set! x1    (list-ref koord 1)   )
                (set! y1    (list-ref koord 2)   )
                (set! x2 (- (list-ref koord 3) 1))
                (set! y2 (- (list-ref koord 4) 1))
                (let loop (
                    (x1 x1) (y1 y1) (x2 x2) (y2 y2)
                    )
                    (set! x1 (+ x1 (if (zero? vertical) 0 1))) 
                    (set! y1 (+ y1 (if (zero? vertical) 1 0))) 
                    (set! x2 (+ x2 (if (zero? vertical) 0 1))) 
                    (set! y2 (+ y2 (if (zero? vertical) 1 0)))                     
                    (set! color1 (car (gimp-image-pick-color image drawable x1 y1 FALSE FALSE 0)))
                    (set! color2 (car (gimp-image-pick-color image drawable x2 y2 FALSE FALSE 0)))
                    (cond
                        ((equal? color1 color2)
                            color1)
                        ((or (>= y2 (car (gimp-image-height image))) (>= x2 (car (gimp-image-width image))))
                            '())
                        (else 
                            (loop x1 y1 x2 y2))
                        )
                    )
                )
            )
        )
    )   
   
   
   
(define (stats m l)
    (cons (/ (length (cdr (member m (reverse l)))) (length l)) (/ (length l)))
    )
   
      
    
(define (script-fu-swap-colors 
    image 
    drawable 
    color1
    color2
    )
    (let* (
            (selection (selection-save image))
            (selection-by-color-save (lambda (color)
                (selection-by-color drawable color 0 selection)
                (if (zero? (car (gimp-selection-is-empty image)))
                    (car (gimp-selection-save image))
                    0
                )))
            (color1-selection (selection-by-color-save color1))
            (color2-selection (selection-by-color-save color2))
            )
        
        (gimp-image-undo-group-start image)
        (gimp-context-push)
        
        (gimp-progress-update 0.25)
        (unless (zero? color1-selection)
            (gimp-selection-load color1-selection)
            (gimp-context-set-foreground color2)
            (gimp-edit-fill drawable FOREGROUND-FILL)
            (gimp-image-remove-channel image color1-selection)    
            )   
        (gimp-progress-update 0.50)         
        (unless (zero? color2-selection)
            (gimp-selection-load color2-selection)
            (gimp-context-set-foreground color1)
            (gimp-edit-fill drawable FOREGROUND-FILL)
            (gimp-image-remove-channel image color2-selection)    
            )        
        (gimp-progress-update 0.75)            
        (unless (zero? selection)
            (gimp-selection-load selection)
            (gimp-image-remove-channel image selection)    
            )
        (gimp-progress-update 1)            
        
        (gimp-context-pop)
        (gimp-image-undo-group-end image)
        (gimp-progress-end)
        (gimp-displays-flush)
        )
    )



(define (script-fu-simutrans-set-grid
    image
    drawable
    tilesize
    resize
    )
    
    (let* (
        (spacing 
            (case tilesize
                ((0) 32)
                ((1) 48)
                ((2) 64)
                ((3) 96)
                ((4) 128)
                ((5) 192)                
                )
            )
        (height     (car (gimp-image-height image)))
        (width      (car (gimp-image-width  image)))
        (new-height 0)
        (new-width  0)
        )
        
        (gimp-image-undo-group-start image)        
        
        (unless (zero? resize)
            (if (zero? (modulo height spacing)) (set! new-height height) (set! new-height (* (+ (quotient height spacing) 1) spacing)))
            (if (zero? (modulo width  spacing)) (set! new-width  width ) (set! new-width  (* (+ (quotient width  spacing) 1) spacing)))
            (if (or (<> height new-height) (<> width new-width)) (gimp-image-resize image new-width new-height 0 0))
            )
        
        (gimp-image-grid-set-spacing image spacing spacing)
        (gimp-image-grid-set-offset image 0 0)    

        (gimp-image-undo-group-end image)
        (gimp-progress-end)
        (gimp-displays-flush)
        )
    )



(define (script-fu-simutrans-export 
    image 
    drawable 
    suffix
    flatten
    threshold
    )
    
    (let* (
            (transparent-color  '(231 255 255))
            
            (image-filename     (car (gimp-image-get-filename image)))
            (base-filename      (car (strbreakup image-filename ".")))
            (extension          ".png")
            
            (buffer-name        (car (gimp-edit-named-copy-visible image "STEXPORT")))
            (export-image       (car (gimp-edit-named-paste-as-new buffer-name)))
            (layer              (car (vector->list (cadr (gimp-image-get-layers export-image)))))
                        
            )
        
        (gimp-context-set-background transparent-color)      
          
        (gimp-context-push)
        (gimp-image-undo-disable export-image)
        (gimp-buffer-delete buffer-name)
        
        (plug-in-threshold-alpha 
            RUN-NONINTERACTIVE 
            export-image 
            layer 
            threshold
            )
            
        (unless (zero? flatten)
            (gimp-layer-flatten layer)
            )
        
        (file-png-save-defaults 
            RUN-NONINTERACTIVE 
            export-image 
            layer 
            (string-append base-filename suffix extension)
            (string-append base-filename suffix extension) 
            )
        
        (gimp-image-delete export-image)   
                
        (gimp-context-pop)     
        (gimp-progress-end)
        (gimp-displays-flush)
            
        )
    
    )
    
    

(define (script-fu-simutrans-special-colors-helper 
    image 
    drawable 
    operation 
    non-darkening-greys
    window-colors
    primary-player-colors
    secondary-player-colors
    lights
    transparent-color
    layers-option
    select-mode
    threshold
    lookup-file
    )

    (let* (      
            (OP-SELECT  0)    
            (OP-REMOVE  1)                
            (OP-REPAIR  2)                
            (OP-LIGHTEN 3)    
            (OP-DARKEN  4)    
            (OP-LOOKUP  5)
            
            (selection (selection-save image))
            (layer-list 
                (cond 
                    ((or 
                        (= layers-option 0)   ;This layer
                        (= layers-option 2))  ;Sample merged (only select)
                        (list drawable)) 
                    ((= layers-option 1)      ;All layers
                        (vector->list (cadr (gimp-image-get-layers image)))) 
                    ))
            
            (non-darkening-greys-list 
                '((107 107 107)  
                  (155 155 155) 
                  (179 179 179) 
                  (201 201 201) 
                  (223 223 223)))
            (window-colors-list 
                '(( 77  77  77)
                  ( 87 101 111)  
                  (193 177 209)))
            (primary-player-colors-list 
                '(( 36  75 103)  
                  ( 57  94 124) 
                  ( 76 113 145)
                  ( 96 132 167)
                  (116 151 189)
                  (136 171 211)
                  (156 190 233)
                  (176 210 255)))
            (secondary-player-colors-list 
                '((123  88   3)  
                  (142 111   4) 
                  (161 134   5)
                  (180 157   7)
                  (198 180   8)
                  (217 203  10)
                  (236 226  11)
                  (255 249  13)))
            (lights-list 
                '((127 155 241) 
                  (255 255  83) 
                  (255  33  29) 
                  (  1 221   1)
                  (227 227 255)
                  (255   1 127)
                  (  1   1 255)))
            (transparent-color-list 
                '((231 255 255)))
                
            (color-set-list   '())
            
            (layer-stats      0)
            (color-set-stats  0)
            (color-stats      0)
            
            (lookup-image     0)
            (lookup-vertical  0)
            )
                
        (gimp-image-undo-group-start image)
        (gimp-context-push)
        
        ;If Sample merged, force Select operation
        (if (= layers-option 2)
            (set! operation OP-SELECT)
            )
            
        ;Add selected special color sets to the list
        (unless (zero? non-darkening-greys    ) (set! color-set-list (cons non-darkening-greys-list      color-set-list)))
        (unless (zero? window-colors          ) (set! color-set-list (cons window-colors-list            color-set-list)))
        (unless (zero? primary-player-colors  ) (set! color-set-list (cons primary-player-colors-list    color-set-list)))
        (unless (zero? secondary-player-colors) (set! color-set-list (cons secondary-player-colors-list  color-set-list)))
        (cond 
            ((= operation OP-DARKEN)) ;do nothing
            ((= operation OP-LIGHTEN) ;reverse color sets from lighter to darker
                (set! color-set-list (map reverse color-set-list)))
            (else                     ;add additional color sets to the list
                (unless (zero? lights           ) (set! color-set-list (cons lights-list                 color-set-list)))
                (unless (zero? transparent-color) (set! color-set-list (cons transparent-color-list      color-set-list)))
                )
            )    
        
        ;Special initialization for some operations
        (cond 
            ;if Select operation, select none
            ((= operation OP-SELECT) 
                (gimp-selection-none image))
            ;if Lookup operation, open lookup image and sets orientation
            ((= operation OP-LOOKUP) 
                (set! lookup-image (car (gimp-file-load RUN-NONINTERACTIVE lookup-file lookup-file)))
                (set! lookup-vertical (if (> (car (gimp-image-height lookup-image)) (car (gimp-image-width lookup-image))) TRUE FALSE)))
            )
              
        ;Repeat for each layer
        (for-each
            (lambda (this-layer)
                (set! layer-stats (stats this-layer layer-list))
        
                ;Repeat for each selected color set
                (for-each 
                    (lambda (this-color-set)
                        (set! color-set-stats (stats this-color-set color-set-list))                    
                
                        ;Repeat for each color in the set    
                        (for-each 
                            (lambda (this-color)
                                (set! color-stats (stats this-color this-color-set))
                            
                                ;Perform chosen operation
                                (cond
                                    
                                    ;Select operation
                                    ((= operation OP-SELECT) 
                                        (gimp-by-color-select this-layer this-color 0 CHANNEL-OP-ADD FALSE FALSE 0 
                                            (if (= layers-option 2) TRUE FALSE) ;Sample merged
                                            )
                                        )
                                    
                                    ;Remove operation
                                    ((= operation OP-REMOVE) 
                                        (selection-by-color this-layer this-color 0 selection)
                                        (selection-fill this-layer 
                                            (map 
                                                (lambda (n) (if (< n 128) (+ n 1) (- n 1))) 
                                                this-color
                                                ))
                                        )
                                    
                                    ;Repair operation   
                                    ((= operation OP-REPAIR)  
                                        (selection-by-color this-layer this-color threshold selection)
                                        (selection-fill this-layer this-color)
                                        )
                                    
                                    ;Lighten/Darken operation   
                                    ((or (= operation OP-LIGHTEN) (= operation OP-DARKEN)) 
                                        (let* ((prev-color (cdr (member this-color (reverse this-color-set)))))
                                            (unless (null? prev-color)
                                                (selection-by-color this-layer this-color 0 selection)
                                                (selection-fill this-layer (car prev-color))
                                                )
                                            )
                                        )
                                        
                                    ;Lookup operation   
                                    ((= operation OP-LOOKUP)
                                        (let* ((lu-color (color-lookup lookup-image this-color lookup-vertical)))
                                            (unless (null? lu-color)
                                                (selection-by-color this-layer lu-color 0 selection)
                                                (selection-fill this-layer this-color)
                                                )
                                            )
                                        )
                                        
                                    ) ;end of case operation
                                
                                ;Update progress bar
                                (gimp-progress-update 
                                    (+     (car layer-stats) 
                                        (* (car color-set-stats) (cdr layer-stats)                       )
                                        (* (car color-stats)     (cdr layer-stats) (cdr color-set-stats) ))
                                    )   
                                
                                ) ;end of lambda (this-color)
                                this-color-set
                            ) ;end of for-each
                            
                        ) ;end of lambda (this-color-set)   
                        color-set-list
                    ) ;end of for-each    
                    
                ) ;end of lambda (this-layer)
                layer-list
            ) ;end of for-each
            
        ;Update progress bar
        (gimp-progress-update 1)    
        
        ;Special termination for some operations
        (cond 
            ;if Lookup operation, close lookup image
            ((= operation OP-LOOKUP) 
                (gimp-image-delete lookup-image))
            )
            
        ;Reset or mask selection        
        (if (zero? selection)
        
            ;No initial selection
            (unless (= operation OP-SELECT) 
                (gimp-selection-none image)
                )
        
            ;Existing selection
            (begin
                (if (= operation OP-SELECT)
                    (case select-mode
                        ((0)) ;do nothing                                             ;Replace current selection
                        ((1) (gimp-selection-combine selection CHANNEL-OP-INTERSECT)) ;Only in current selection
                        ((2) (gimp-selection-combine selection CHANNEL-OP-ADD))       ;Add to current selection
                        ((3) (gimp-selection-invert image)                            ;Subtract from current selection
                             (gimp-selection-combine selection CHANNEL-OP-INTERSECT))
                        )
                    (gimp-selection-load selection)
                    )
                (gimp-image-remove-channel image selection)    
                )
            )
                            
            (gimp-context-pop)
            (gimp-image-undo-group-end image)
            
            (gimp-progress-end)
            (gimp-displays-flush)
        )
    )

(script-fu-register "script-fu-simutrans-special-colors-helper"
    "Simutrans Special Colors _Helper..."
    "Select, remove or repair Simutrans special colors"
    "Fabio Gonella"
    "Fabio Gonella"
    "March 2012"
    "RGB*"
    SF-IMAGE      "Image"    0
    SF-DRAWABLE   "Drawable" 0
    
    SF-OPTION     
        "_Operation to perform" 
        '(
            "Select special colors" 
            "Remove special colors" 
            "Repair special colors"
            "Lighten special colors"
            "Darken special colors"
            "Convert to special colors"
            )
    
    SF-TOGGLE     "_Non-darkening greys"                  TRUE
    SF-TOGGLE     "_Windows"                              TRUE
    SF-TOGGLE     "Player colors (_Primary)"              TRUE
    SF-TOGGLE     "Player colors (_Secondary)"            TRUE
    SF-TOGGLE     "_Lights (except lighten/darken)"       FALSE
    SF-TOGGLE     "_Transparent (except lighten/darken)"  FALSE
    
    SF-OPTION     
        "_Apply to" 
        '(
            "This layer"
            "All layers" 
            "Sample merged (only select)"
            )  
              
    SF-OPTION     
        "Selection _mode (only select)" 
        '(
            "Replace current selection"
            "Only in current selection" 
            "Add to current selection" 
            "Subtract from current selection"
            )
            
    SF-ADJUSTMENT "Th_reshold (only repair)"              '(15 0 255 1 10 1 0)
    
    SF-FILENAME "Look_up image (only convert)"            (string-append "" gimp-data-directory "/scripts/images/TTD.png")
    )

(script-fu-register "script-fu-simutrans-export"
    "_Export with transparent background..."
    "Export to PNG adding a transparent special color background"
    "Fabio Gonella"
    "Fabio Gonella"
    "March 2012"
    "RGB*"
    SF-IMAGE      "Image"       0
    SF-DRAWABLE   "Drawable"    0
    SF-STRING     "_Custom suffix (e.g. -01)" ""
    SF-TOGGLE     "_Flatten Alpha Channel" TRUE
    SF-ADJUSTMENT "Alpha _Threshold" '(127 0 255 1 10 1 0)    
    )

(script-fu-register "script-fu-simutrans-set-grid"
    "Set tiles _grid..."
    "Set grid for chosen tileset size"
    "Fabio Gonella"
    "Fabio Gonella"
    "March 2012"
    "RGB*"
    SF-IMAGE      "Image"       0
    SF-DRAWABLE   "Drawable"    0
    SF-OPTION     "Tile _Size"  '("32" "48" "64" "96" "128" "192")    
    SF-TOGGLE     "Resize"      FALSE
    )

(script-fu-register "script-fu-swap-colors"
    "_Swap two colors..."
    "Swap two colors in selection"
    "Fabio Gonella"
    "Fabio Gonella"
    "March 2012"
    "RGB*"
    SF-IMAGE      "Image"    0
    SF-DRAWABLE   "Drawable" 0
    SF-COLOR      "Color _1" (car (gimp-context-get-foreground))
    SF-COLOR      "Color _2" (car (gimp-context-get-background))    
    )
    
(script-fu-menu-register "script-fu-simutrans-special-colors-helper"
    "<Image>/Simutrans" )

(script-fu-menu-register "script-fu-simutrans-export"
    "<Image>/Simutrans" )

(script-fu-menu-register "script-fu-simutrans-set-grid"
    "<Image>/Simutrans/Tile Tools" )

(script-fu-menu-register "script-fu-swap-colors"
    "<Image>/Simutrans/Color Tools" )
